local M = {}
local tconcat = table.concat
local tremove = table.remove
local nexttoken
local tag = [["'</>=]]

-- Token表，用于处理不同的HTML字符
local T = {
    [tag:byte(1)] = function(str, start) -- 处理双引号
        local e = str:find('[^\\]"', start)
        return str:sub(start+1, e), e + 2
    end,
    [tag:byte(2)] = function(str, start) -- 处理单引号
        local e = str:find("[^\\]'", start)
        return str:sub(start+1, e), e + 2
    end,
    [tag:byte(3)] = function(str, start) -- 处理开标签符号 (<)
        local n = str:byte(start + 1)
        local s = "!/-"
        if n == s:byte(1) then -- 处理注释
            if str:byte(start + 2) == s:byte(3) then -- 注释结束符 (-->)
                local _, e = str:find("-->", start + 4, true)
                return nexttoken(str, e + 1)
            else
                local e = str:find(">", start + 1)
                return nexttoken(str, e + 1)
            end
        elseif n == s:byte(2) then -- 处理关闭标签
            return "</", start + 2
        end
        return "<", start + 1
    end,
    [tag:byte(4)] = function(str, start) -- 处理自闭合标签 (/>)
        if str:byte(start + 1) == tag:byte(5) then
            return "/>", start + 2
        end
        return "/", start + 1
    end,
    [tag:byte(5)] = function(str, start) -- 处理结束标签 (>)
        return ">", start + 1
    end,
    [tag:byte(6)] = function(str, start) -- 处理等号 (=)
        return "=", start + 1
    end,
}

-- 处理普通文本的函数
local function plain(str, start)
    local e = str:find("[%s><=/]", start)
    assert(e > start)
    return str:sub(start, e - 1), e
end

-- 处理文本内容
local function nexttext(str, start)
    local start = str:find('[^%s]', start)
    if not start then
        return nil
    end
    local e = str:find("<", start)
    assert(e > start)
    return str:sub(start, e - 1), e
end

-- 读取下一个token
function nexttoken(str, start)
    local start = str:find('[^%s]', start)
    if not start then
        return nil
    end
    local n = str:byte(start)
    local func = T[n]
    if not func then -- 普通文本处理
        func = plain
    end
    return func(str, start)
end

-- 节点对象定义
local node = {
    -- 获取节点的所有子节点内容
    value = function(self)
        local tbl = {}
        -- 遍历所有子节点
        for _, v in pairs(self.child) do
            if type(v) == "string" then
                -- 如果是字符串，添加到 buffer 中
                tbl[#tbl + 1] = v
            elseif type(v) == "table" and v.value then
                -- 如果是表格，并且有 value 方法，递归调用
                tbl[#tbl + 1] = v:value()  -- 递归获取子节点的内容
            else
				v:value(tbl)
			end
        end
        -- 返回拼接的文本内容
        return tconcat(tbl)
    end,
    
    -- 获取节点的文本内容
    text = function(self)
        local tbl = {}
        self:value(tbl)
        return tconcat(tbl)
    end,

    -- 节点匹配选择器条件的函数
    match = function(self, cond)
        local match = true
        local class = cond.class

        -- 处理类选择器
        if class then
            match = false
            for _, v in pairs(self.class) do
                if v == class then
                    match = true
                    break
                end
            end
        end

        -- 处理ID选择器
        local id = cond.id
        if match and id and self.attr["id"] ~= id then
            match = false
        end

        -- 处理name选择器
        local name = cond.name
        if match and name and self.name ~= name then
            match = false
        end

        -- 处理属性选择器
        if match and cond.attr then
            for key, value in pairs(cond.attr) do
                if self.attr[key] ~= value then
                    match = false
                    break
                end
            end
        end

        return match
    end,

    -- 用于选择符合条件的节点
    select = function(self, method)
        local out = {}
        local childs = {}
        local cond = {
            name = nil,
            id = nil,
            class = nil,
            attr = {}  -- 属性选择器字段
        }
        local pattern = ".#"

        -- 解析选择器
        for k in method:gmatch("([.#]-[^.#%s]+)") do
            local n = k:byte(1)
            if n == pattern:byte(1) then -- 类选择器
                cond.class = k:sub(2)
            elseif n == pattern:byte(2) then -- ID选择器
                cond.id = k:sub(2)
            elseif k:find("%[") then -- 属性选择器
                local attr_key, attr_value = k:match("%[(%w+)=\"([^\"]+)\"%]")
                if attr_key and attr_value then
                    cond.attr[attr_key] = attr_value
                end
            else
                -- 标签名选择器
                cond.name = k
            end
        end

        local nodes = {self}

        -- 遍历节点树
        while true do
            local n = tremove(nodes, 1)
            if not n then
                break
            end
            for _, v in pairs(n.child) do
                if type(v) == "table" then
                    nodes[#nodes + 1] = v
                end
            end
            if n:match(cond) then
                out[#out + 1] = n
            end
        end
        
        return out
    end,

    -- 用于多层级选择器
    selectn = function(self, method, level)
        local item = self
        for i = 1, level - 1 do
            item = item:select(method)[1]
        end
        return item:select(method)
    end
}

local nodemt = {__index = node}

-- 定义自闭合标签
local selfclose = {
    ["META"] = true,
    ["LINK"] = true,
    ["INPUT"] = true,
    ["IMG"] = true,
    ["BR"] = true,
    ["HR"] = true,
}

-- HTML转义字符的处理
local html_unescape = {
    ['quot'] = '"',
    ['amp'] = '&',
    ['lt'] = '<',
    ['gt'] = '>',
    ['nbsp'] = ' ',
}

-- 将HTML转义字符转换回普通字符
function htmlunescape(html)
    html = string.gsub(html, "&#(%d+);", function(s)
        return utf8.char(tonumber(s))
    end)
    html = string.gsub(html, "&(%a+);", html_unescape)
    return html
end

-- 解析单个HTML节点
local function parsenode(str, start)
    local tk, name, back
    tk, start = nexttoken(str, start)
    if not tk then
        return
    end
    assert(tk == "<")
    name, start = nexttoken(str, start)

    local attr = {}
    local class = {}
    local child = {}
    
    -- 创建节点对象
    local obj = setmetatable({
        name = name,
        attr = attr,
        class = class,
        child = child,
    }, nodemt)

    -- 解析标签属性
    while true do
        local x, val
        tk, start = nexttoken(str, start)
        if tk == ">" or tk == "/>" then
            break
        end
        back = start
        x, start = nexttoken(str, start)
        if x == "=" then
            local v
            v, start = nexttoken(str, start)
            if tk == "class" then -- 处理class属性
                for c in v:gmatch("[^%s]+") do
                    class[#class + 1] = c
                end
                local a = attr[tk]
                if a then
                    v = a .. " " .. v
                end
                attr[tk] = v
            else
                attr[tk] = v
            end
        else
            attr[tk] = true
            start = back
        end
    end

    -- 处理自闭合标签
    if tk == "/>" or selfclose[name:upper()] then
        return obj, start
    end
    assert(tk == ">")

        -- 处理script标签内容
    if name == "script" then
        local s, e = str:find("</script>", start)
        child[#child + 1] = str:sub(start, s - 1)
        return obj, e + 1
    elseif name == "style" then
        -- 处理style标签内容
        local s, e = str:find("</style>", start)
        child[#child + 1] = str:sub(start, s - 1)
        return obj, e + 1
    end

    -- 处理子节点
    while true do
        local tk
        back = start
        tk, start = nexttoken(str, start)
        if not tk then
            break
        elseif tk == "<" then -- 打开新标签
            child[#child + 1], start = parsenode(str, back)
        elseif tk == "</" then -- 关闭标签
            name, start = nexttoken(str, start)
            tk, start = nexttoken(str, start)
            assert(tk == ">")
            if name ~= obj.name then
                start = back
            end
            break
        else
            -- 处理标签间的文本
            tk, start = nexttext(str, back)
            local raw = htmlunescape(tk)
            child[#child + 1] = raw
        end
    end

    return obj, start
end

-- 解析HTML字符串，并返回根节点
M.parse = function(str)
    local ok, node = pcall(parsenode, str, 1)
    if ok then
        return node
    else
        return nil, node
    end
end

return M