local ffi = require "ffi"

-- 定义 iconv 的 C 函数接口
ffi.cdef[[
typedef void* iconv_t;

iconv_t iconv_open(const char* tocode, const char* fromcode);
int iconv(iconv_t cd, char** inbuf, size_t* inbytesleft, char** outbuf, size_t* outbytesleft);
int iconv_close(iconv_t cd);
]]

-- 加载 libiconv.so
local libiconv = ffi.load("/usr/local/lor/libs/libiconv.so")

-- 封装 iconv 模块
local iconv = {}

-- 创建一个新的转换实例，模仿 lua-iconv 的接口
function iconv.new(to_encoding, from_encoding)
    local cd = libiconv.iconv_open(to_encoding, from_encoding)
    if cd == ffi.cast("iconv_t", -1) then
        return nil, "Failed to open iconv with to_encoding=" .. to_encoding .. ", from_encoding=" .. from_encoding
    end

    -- 返回封装好的转换实例
    return {
        cd = cd,

        -- iconv 转换方法，模仿 lua-iconv 的 `iconv:iconv()` 方法
        convert = function(self, input_str)
            local input_len = #input_str

            -- 为输出准备足够的空间
            local output_len = input_len * 2  -- 通常输出缓冲区要比输入缓冲区大
            local output_buf = ffi.new("char[?]", output_len)
            local input_buf = ffi.new("char[?]", input_len + 1)  -- 给输入字符串分配内存
            ffi.copy(input_buf, input_str)  -- 把 Lua 字符串复制到 C 字符数组

            local inbuf = ffi.new("char*[1]", input_buf)
            local outbuf = ffi.new("char*[1]", output_buf)
            local inbytesleft = ffi.new("size_t[1]", input_len)
            local outbytesleft = ffi.new("size_t[1]", output_len)

            -- 执行转换
            local result = libiconv.iconv(self.cd, inbuf, inbytesleft, outbuf, outbytesleft)
            if result == -1 then
                return nil, "Conversion failed"
            end

            -- 从输出缓冲区构造结果字符串
            return ffi.string(output_buf, output_len - outbytesleft[0])
        end,

        -- 关闭 iconv
        close = function(self)
            libiconv.iconv_close(self.cd)
        end
    }
end

return iconv